/* 
 * File:   evaluator8.cpp
 * Article: http://www.saboti.nu/writing-an-expression-evaluator/index.html
 *
 * AGPL3
 */
 
/*
 * Left to right parser
 * Parentheses
 * Precedence
 * Power operator
 * Associativity
 * Prefix
 * Suffix
 * AST
 */

#include <cstdlib>
#include <iostream>
#include <cmath>

enum class Operator
{
	Null,
	Addition,
	Subtraction,
	Multiplication,
	Division,
	Power,
	Positive,
	Negative,
	Factorial,
};

/*----------------------------------------------------------------------------*/

class AST
{
public:
	virtual int eval() = 0;
};

bool read_expression(char const *& pos, AST *& node, Operator previous_op = Operator::Null);
bool read_prefix_operator(char const *& pos, Operator & op);
bool compute_unary(Operator op, int arg, int & result);
bool compute_binary(Operator op, int lhs, int rhs, int & result);

class ASTNumber : public AST
{
private:
	int d_number;
public:
	ASTNumber(int number) : d_number(number)
	{}
	int eval() override
	{
		return d_number;
	}
};

class ASTOperator : public AST
{
protected:
	Operator const d_operator;
public:
	ASTOperator(Operator op) : d_operator(op)
	{}
};
 
class ASTBinary : public ASTOperator
{
private:
	AST * d_lhs = nullptr;
	AST * d_rhs = nullptr;
public:
	ASTBinary(Operator op, AST * lhs, AST * rhs) : ASTOperator(op), d_lhs(lhs), d_rhs(rhs)
	{}
	int eval() override
	{
		int result;
		if(!compute_binary(d_operator, d_lhs->eval(), d_rhs->eval(), result))
		{
			std::abort();
		}
		return result;
	}
};

class ASTUnary : public ASTOperator
{
private:
	AST * d_arg = nullptr;
public:
	ASTUnary(Operator op, AST * arg) : ASTOperator(op), d_arg(arg)
	{}
	int eval() override
	{
		int result;
		if(!compute_unary(d_operator, d_arg->eval(), result))
		{
			std::abort();
		}
		return result;
	}
};

/*----------------------------------------------------------------------------*/

enum class Associativity
{
	Left,
	Right,
};

bool is_digit(char const c)
{
	return c >= '0' && c <= '9';
}

bool read_number(char const *& pos, AST *& num)
{
	if(*pos == '(')
	{
		++pos;
		if(!read_expression(pos, num) || *pos != ')')
		{
			return false;
		}
		++pos;
		return true;
	}
	
	Operator op;
	if(read_prefix_operator(pos, op))
	{
		if(!read_expression(pos, num, op))
		{
			return false;
		}
		num = new ASTUnary(op, num);
		return true;
	}
	
	int number = 0;
	if(is_digit(*pos))
	{
		do
		{
			number *= 10;
			number += *pos - '0';
			++pos;
		} while(is_digit(*pos));
		num = new ASTNumber(number);
		return true;
	}
	return false;
}

bool read_operator(char const *& pos, Operator & op)
{
	switch(*pos)
	{
		case '+':
			op = Operator::Addition;
			break;
		case '-':
			op = Operator::Subtraction;
			break;
		case '*':
			op = Operator::Multiplication;
			break;
		case '/':
			op = Operator::Division;
			break;
		case '^':
			op = Operator::Power;
			break;
		case '!':
			op = Operator::Factorial;
			break;
		default:
			return false;
	}
	++pos;
	return true;
}

bool read_prefix_operator(char const *& pos, Operator & op)
{
	switch(*pos) 
	{
		case '-':
			op = Operator::Negative;
			break;
		case '+':
			op = Operator::Positive;
			break;
		default:
			return false;
	}
	++pos;
	return true;
}

int p(Operator op)
{
	switch(op)
	{
		case Operator::Addition:
		case Operator::Subtraction:
			return 10;
		case Operator::Multiplication:
		case Operator::Division:
			return 20;
		case Operator::Negative:
		case Operator::Positive:
			return 25;
		case Operator::Factorial:
			return 27;
		case Operator::Power:
			return 30;
		default:
			return 0;
	}
}

Associativity a(Operator op)
{
	switch(op)
	{
		case Operator::Power:
			return Associativity::Right;
		default:
			return Associativity::Left;
	}
}

bool is_suffix_operator(Operator op)
{
	switch(op)
	{
		case Operator::Factorial:
			return true;
	}
	return false;
}

bool compute_binary(Operator op, int lhs, int rhs, int & result)
{
	switch(op)
	{
		case Operator::Addition:
			result = lhs + rhs;
			break;
		case Operator::Subtraction:
			result = lhs - rhs;
			break;
		case Operator::Multiplication:
			result = lhs * rhs;
			break;
		case Operator::Division:
			result = lhs / rhs;
			break;
		case Operator::Power:
			result = std::pow(lhs, rhs);
			break;
		default:
			return false;
	}
	return true;
}

bool compute_unary(Operator op, int arg, int & result)
{
	switch(op)
	{
		case Operator::Negative:
			result = arg * -1;
			break;
		case Operator::Positive:
			result = arg * 1;
			break;
		case Operator::Factorial:
			if(arg < 0)
			{
				return false;
			}
			result = 1;
			for( ; arg > 1; --arg)
			{
				result *= arg;
			}
			break;
		default:
			return false;
	}
	return true;
}

bool read_expression(char const *& pos, AST *& node, Operator previous_op)
{
	if(read_number(pos, node))
	{
		Operator op;
		while(read_operator(pos, op))
		{
			/*
			 * Precedence check.
			 */
			bool op_less_previous = p(op) < p(previous_op);
			if(op_less_previous)
			{
				--pos;
				break;
			}
			/*
			 * Suffix
			 */
			if(is_suffix_operator(op))
			{
				node = new ASTUnary(op, node);
				continue;
			}
			/*
			 * Associativity check.
			 */
			bool op_equal_previous = p(op) == p(previous_op);
			bool op_left_associativity = a(op) == Associativity::Left;
			if(op_equal_previous && op_left_associativity)
			{
				--pos;
				break;
			}
			AST * rhs;
			if(!read_expression(pos, rhs, op))
			{
				return false;
			}
			node = new ASTBinary(op, node, rhs);
		}
		return true;
	}
	return false;
}

bool calculate(char const *& pos, AST *& node)
{
	if(!pos)
	{
		return false;
	}
	return read_expression(pos, node) && !*pos;
}

int main(int argc, char const ** argv) 
{
	
	if(!argv[1])
	{
		std::cout << "You must invoke this program with an expression\n";
		return 1;
	}
	
	char const * text = argv[1];
	AST * result;
	
	if(calculate(text, result))
	{
		std::cout << result->eval() << std::endl;
	}
	else
	{
		std::cout << "There seems to be an error here: '" << text << "'\n";
		return 1; 
	}
	
	return 0;
}

